from abc import ABC, abstractmethod
from typing_extensions import AsyncIterable, Sequence
from dataclasses import dataclass
from decimal import Decimal
from datetime import datetime
from trading_sdk.types import Side


@dataclass
class Trade:
  @dataclass
  class Fee:
    asset: str
    amount: Decimal

  id: str
  price: Decimal
  qty: Decimal
  time: datetime
  side: Side
  fee: Fee | None = None
  maker: bool | None = None

class MyTrades(ABC):
  @abstractmethod
  def my_trades(
    self, base: str, quote: str, *,
    start: datetime | None = None, end: datetime | None = None
  ) -> AsyncIterable[Sequence[Trade]]:
    """Fetch trades (from your account) on a given symbol. Automatically paginates if needed.
    
    - `base`: The base asset, e.g. `BTC`.
    - `quote`: The quote asset, e.g. `USDT`.
    - `start`: The start time to query. If given, only trades after this time will be returned.
    - `end`: The end time to query. If given, only trades before this time will be returned.
    """
    ...

  async def my_trades_seq(
    self, base: str, quote: str, *,
    start: datetime | None = None, end: datetime | None = None
  ) -> Sequence[Trade]:
    """Fetch trades (from your account) on a given symbol. Automatically paginates if needed.
    
    - `base`: The base asset, e.g. `BTC`.
    - `quote`: The quote asset, e.g. `USDT`.
    - `start`: The start time to query. If given, only trades after this time will be returned.
    - `end`: The end time to query. If given, only trades before this time will be returned.
    """
    trades: list[Trade] = []
    async for batch in self.my_trades(base, quote, start=start, end=end):
      trades.extend(batch)
    return trades


from abc import ABC, abstractmethod
from trading_sdk.types import Num

class EditOrder(ABC):
  @abstractmethod
  async def edit_order(self, base: str, quote: str, *, id: str, qty: Num) -> str:
    """Edit an existing order.
    
    - `base`: The base asset, e.g. `BTC`.
    - `quote`: The quote asset, e.g. `USDT`.
    - `id`: The ID of the order to edit.
    - `quantity`: The new quantity of the order.

    Returns the new ID of the order.
    """
    ...


from abc import ABC, abstractmethod
from typing_extensions import AsyncIterable, Sequence
from dataclasses import dataclass
from decimal import Decimal
from datetime import datetime
from trading_sdk.types import Num

@dataclass
class Reward:
  asset: Decimal
  qty: Decimal
  time: datetime

class RewardHistory(ABC):
  @abstractmethod
  async def reward_history(
    self, *, start: datetime | None = None,
    end: datetime | None = None
  ) -> AsyncIterable[Sequence[Reward]]:
    """Fetch earn rewards.
    
    - `start`: if given, retrieves rewards after this time.
    - `end`: if given, retrieves rewards before this time.
    """
    ...


from abc import ABC, abstractmethod
from trading_sdk.spot.user_data.query_order import OrderState

class CancelOrder(ABC):
  @abstractmethod
  async def cancel_order(self, base: str, quote: str, *, id: str) -> OrderState:
    """Cancel an order.
    
    - `base`: The base asset, e.g. `BTC`.
    - `quote`: The quote asset, e.g. `USDT`.
    - `id`: The ID of the order to cancel.

    Returns the state of the order.
    """
    ...


from abc import ABC, abstractmethod
from typing_extensions import Sequence
from .query_order import OrderState

class OpenOrders(ABC):
  @abstractmethod
  async def open_orders(self, base: str, quote: str) -> Sequence[OrderState]:
    """Fetch currently open orders (of your account) on a given symbol.
    
    - `base`: The base asset, e.g. `BTC`.
    - `quote`: The quote asset, e.g. `USDT`.
    """
    ...


from abc import ABC, abstractmethod
from typing_extensions import Literal, AsyncIterable, Sequence
from dataclasses import dataclass
from datetime import datetime
from decimal import Decimal

@dataclass
class AggTrade:
  price: Decimal
  qty: Decimal
  """Base asset quantity."""
  time: datetime
  """Time of the transaction."""
  maker: Literal['buyer', 'seller']
  """Which side was the maker."""

  
class AggTrades(ABC):
  @abstractmethod
  async def trades(
    self, base: str, quote: str, *,
    start: datetime | None = None, end: datetime | None = None
  ) -> AsyncIterable[Sequence[AggTrade]]:
    """Fetch aggregate trades for the given symbol. Automatically paginates if needed.
    
    - `base`: The base asset, e.g. `BTC`.
    - `quote`: The quote asset, e.g. `USDT`.
    - `start`: if given, retrieves trades after this time.
    - `end`: if given, retrieves trades before this time.

    ### Ordering

    - If `start` is given, trades are ordered forwards by time.
    - Otherwise, ordered backwards by time.
    """
    ...
  


from dataclasses import dataclass
from abc import ABC, abstractmethod
from trading_sdk.types import Network

@dataclass
class Address:
  address: str
  memo: str | None = None

class DepositAddress(ABC):
  @abstractmethod
  async def deposit_address(self, asset: str, *, network: Network) -> Address:
    """Get the deposit address for an asset.
    
    - `asset`: The asset to get the deposit address for.
    - `network`: The network to get the deposit address for.
    """
    ...


from typing_extensions import Sequence
import asyncio
from .place_order import Order, PlaceOrder

class PlaceOrders(PlaceOrder):
  async def place_orders(self, base: str, quote: str, orders: Sequence[Order]) -> Sequence[str]:
    """Place multiple orders on a given symbol.
    
    - `base`: The base asset, e.g. `BTC`.
    - `quote`: The quote asset, e.g. `USDT`.
    - `orders`: The orders to place.

    Returns the order IDs.
    """
    return await asyncio.gather(*[self.place_order(base, quote, order) for order in orders])


from abc import ABC, abstractmethod
from trading_sdk.types import Num

class Subscribe(ABC):
  @abstractmethod
  async def subscribe(self, product: str, *, amount: Num):
    """Subscribe to an earn product.
    
    - `product`: The product to subscribe to.
    - `amount`: The amount to subscribe to.
    """
    ...


from .wallet import Wallet
from .spot import Spot
from .futures import Futures
from .earn import Earn
from .types import Side, TimeInForce, fmt_num, Error, NetworkError, ValidationError, UserError, AuthError

__all__ = [
  'Wallet',
  'Spot',
  'Futures',
  'Earn',
  'Side',
  'TimeInForce',
  'fmt_num',
  'Error', 'NetworkError', 'ValidationError', 'UserError', 'AuthError',
]


from abc import ABC, abstractmethod
from typing_extensions import Sequence
from dataclasses import dataclass
from decimal import Decimal
from trading_sdk.types import Network

@dataclass
class WithdrawalMethod:

  @dataclass
  class Fee:
    asset: str
    amount: Decimal

  network: Network
  fee: Fee | None = None
  contract_address: str | None = None

class WithdrawalMethods(ABC):
  @abstractmethod
  async def withdrawal_methods(self, asset: str) -> Sequence[WithdrawalMethod]:
    """Get the withdrawal methods for an asset.
    
    - `asset`: The asset to get the withdrawal methods for.
    """
    ...


from abc import ABC, abstractmethod
from typing_extensions import AsyncIterable, Sequence, Literal
from dataclasses import dataclass
from decimal import Decimal
from datetime import datetime

PositionType = Literal['LONG', 'SHORT']

@dataclass
class Funding:
  rate: Decimal
  funding: Decimal
  """Funding paid (if negative) or earned (if positive)."""
  time: datetime
  position_type: PositionType

class FundingRateHistory(ABC):
  @abstractmethod
  def funding_rate_history(
    self, base: str, quote: str, *, start: datetime | None = None,
    end: datetime | None = None,
  ) -> AsyncIterable[Sequence[Funding]]:
    """Fetch funding rate history for a given symbol.

    - `base`: The base asset, e.g. `BTC`.
    - `quote`: The quote asset, e.g. `USDT`.
    - `start`: if given, retrieves funding rates after this time.
    - `end`: if given, retrieves funding rates before this time.
    """
    ...


from abc import ABC, abstractmethod
from trading_sdk.types import Num, Network

class Withdraw(ABC):
  @abstractmethod
  async def withdraw(
    self, asset: str, *, address: str, amount: Num,
    network: Network, memo: str | None = None,
    contract_address: str | None = None,
  ):
    """Withdraw funds.
    
    - `asset`: The asset to withdraw.
    - `address`: The address to withdraw to.
    - `amount`: The amount to withdraw.
    - `network`: The network to withdraw to.
    - `memo`: The memo to withdraw to (for some networks).
    - `contract_address`: The contract address to withdraw to. You can use it to make sure it's the token you expect.
    """
    ...


from abc import ABC, abstractmethod
from trading_sdk.types import Num

class Redeem(ABC):
  @abstractmethod
  async def redeem(self, product: str, *, amount: Num):
    """Redeem from an earn product.
    
    - `product`: The product to redeem.
    - `amount`: The amount to redeem.
    """
    ...


from abc import ABC, abstractmethod
from typing_extensions import Mapping, TypeVar
from dataclasses import dataclass
from decimal import Decimal
import asyncio

S = TypeVar('S', bound=str, default=str)

@dataclass
class Balance:
  free: Decimal
  locked: Decimal

  @property
  def total(self) -> Decimal:
    return self.free + self.locked

class Balances(ABC):
  @abstractmethod
  async def balance(self, currency: S, /) -> Balance:
    """Get the balance of the given currency."""
    return (await self.balances(currency))[currency]

  @abstractmethod
  async def balances(self, *currencies: S) -> Mapping[S, Balance]:
    """Get the balances of the given currencies."""
    balances = await asyncio.gather(*(self.balance(currency) for currency in currencies))
    return {currency: balance for currency, balance in zip(currencies, balances)}


from abc import ABC, abstractmethod
from typing_extensions import TypedDict, Literal
from trading_sdk.types import Side, Num

class BaseOrder(TypedDict):
  side: Side
  qty: Num
  """Quantity of the order in the base asset."""

class LimitOrder(BaseOrder):
  price: Num
  type: Literal['LIMIT']

class MarketOrder(BaseOrder):
  type: Literal['MARKET']

Order = LimitOrder | MarketOrder

class PlaceOrder(ABC):
  @abstractmethod
  async def place_order(self, base: str, quote: str, order: Order) -> str:
    """Place an order.
    
    - `base`: The base asset, e.g. `BTC`.
    - `quote`: The quote asset, e.g. `USDT`.
    - `order`: The order to place.

    Returns the order ID.
    """
    ...


from abc import ABC, abstractmethod
from typing_extensions import AsyncIterable
from trading_sdk.spot.market_data.depth import Book

class Depth(ABC):
  @abstractmethod
  def depth(self, base: str, quote: str, *, limit: int | None = None) -> AsyncIterable[Book]:
    """Stream of Book snapshots for the given symbol.
    
    - `base`: The base asset, e.g. `BTC`.
    - `quote`: The quote asset, e.g. `USDT`.
    - `limit`: The maximum number of bids/asks to return.
    """


from typing_extensions import Literal, TypeGuard

Network = Literal[
  'BSC',
  'ETH',
  'POL',
  'TRX',
  'ARB',
  'BASE',
  'AVAX',
  'SOL',
  'FTM',
  'OP',
  'TON',
  'SONIC',
  'XTZ',
  'CORE',
  'CELO',
  'NEAR',
  'MNT',
  'UNI',
  'APTOS',
  'SUI',
]

NETWORK_NAMES: dict[Network, str] = {
  'BSC': 'Binance Smart Chain',
  'ETH': 'Ethereum',
  'POL': 'Polygon',
  'TRX': 'Tron',
  'ARB': 'Arbitrum',
  'BASE': 'Base',
  'AVAX': 'Avalanche',
  'SOL': 'Solana',
  'FTM': 'Fantom',
  'OP': 'Optimism',
  'TON': 'The Open Network',
  'SONIC': 'Sonic',
  'XTZ': 'Tezos',
  'CORE': 'Core',
  'CELO': 'Celo',
  'NEAR': 'Near',
  'MNT': 'Mantle',
  'UNI': 'Unichain',
  'APTOS': 'Aptos',
  'SUI': 'Sui',
}

def is_network(s: str) -> TypeGuard[Network]:
  return s in NETWORK_NAMES


from typing_extensions import AsyncIterable, Sequence
from abc import ABC, abstractmethod
from dataclasses import dataclass
from decimal import Decimal
from datetime import datetime
from trading_sdk.types import Network

@dataclass
class Withdrawal:
  @dataclass
  class Fee:
    asset: str
    amount: Decimal

  id: str
  amount: Decimal
  asset: str
  time: datetime
  address: str | None = None
  network: Network | None = None
  fee: Fee | None = None
  memo: str | None = None

class WithdrawalHistory(ABC):
  @abstractmethod
  def withdrawal_history(
    self, *, asset: str | None = None,
    start: datetime,
    end: datetime,
  ) -> AsyncIterable[Sequence[Withdrawal]]:
    """Fetch your withdrawals.
    
    - `asset`: if given, retrieves withdrawals for this asset.
    - `start`: if given, retrieves withdrawals after this time.
    - `end`: if given, retrieves withdrawals before this time.
    """
    ...


from ._market_data import MarketData
from .depth import Depth
from .exchange_info import ExchangeInfo
from .time import Time
from .agg_trades import AggTrades

__all__ = [
  'MarketData',
  'Depth',
  'ExchangeInfo',
  'Time',
  'AggTrades',
]


from abc import ABC, abstractmethod
from typing_extensions import AsyncIterable
from trading_sdk.spot.user_data.my_trades import Trade

class MyTrades(ABC):
  @abstractmethod
  def my_trades(self, base: str, quote: str) -> AsyncIterable[Trade]:
    """Stream of your trades for the given symbol.
    
    - `base`: The base asset, e.g. `BTC`.
    - `quote`: The quote asset, e.g. `USDT`.
    """
    ...


import asyncio
from typing_extensions import Sequence
from .query_order import QueryOrder, OrderState

class QueryOrders(QueryOrder):
  async def query_orders(self, base: str, quote: str, *, ids: Sequence[str]) -> Sequence[OrderState]:
    """Query multiple orders by symbol and ID.
    
    - `base`: The base asset, e.g. `BTC`.
    - `quote`: The quote asset, e.g. `USDT`.
    - `ids`: The IDs of the orders to query.

    Returns the states of the orders.
    """
    return await asyncio.gather(*[self.query_order(base, quote, id=id) for id in ids])


from ._wallet import Wallet
from .deposit_address import DepositAddress
from .withdraw import Withdraw
from .withdrawal_methods import WithdrawalMethods
from .deposit_history import DepositHistory
from .withdrawal_history import WithdrawalHistory

__all__ = [
  'Wallet', 'DepositAddress', 'Withdraw', 'WithdrawalMethods',
  'DepositHistory', 'WithdrawalHistory',
]


from .exc import Error, NetworkError, ValidationError, UserError, AuthError, ApiError
from .misc import Side, TimeInForce, Num, fmt_num
from .networks import Network, NETWORK_NAMES, is_network

__all__ = [
  'Error', 'NetworkError', 'ValidationError', 'UserError', 'AuthError', 'ApiError',
  'Side', 'TimeInForce', 'Num', 'fmt_num',
  'Network', 'NETWORK_NAMES', 'is_network',
]


from .deposit_address import DepositAddress
from .withdraw import Withdraw
from .withdrawal_methods import WithdrawalMethods
from .deposit_history import DepositHistory
from .withdrawal_history import WithdrawalHistory

class Wallet(DepositAddress, Withdraw, WithdrawalMethods, DepositHistory, WithdrawalHistory):
  ...